@@ -32,7 +32,19 @@
 
 
 
+typedef struct _CategoryItem CategoryItem;
+
+
+
 static void               xfce_appfinder_category_model_tree_model_init       (GtkTreeModelIface        *iface);
+static void               xfce_appfinder_category_model_get_property          (GObject                  *object,
+                                                                               guint                     prop_id,
+                                                                               GValue                   *value,
+                                                                               GParamSpec               *pspec);
+static void               xfce_appfinder_category_model_set_property          (GObject                  *object,
+                                                                               guint                     prop_id,
+                                                                               const GValue             *value,
+                                                                               GParamSpec               *pspec);
 static void               xfce_appfinder_category_model_finalize              (GObject                  *object);
 static GtkTreeModelFlags  xfce_appfinder_category_model_get_flags             (GtkTreeModel             *tree_model);
 static gint               xfce_appfinder_category_model_get_n_columns         (GtkTreeModel             *tree_model);
@@ -63,6 +75,7 @@ static gboolean           xfce_appfinder_category_model_iter_nth_child        (G
 static gboolean           xfce_appfinder_category_model_iter_parent           (GtkTreeModel             *tree_model,
                                                                                GtkTreeIter              *iter,
                                                                                GtkTreeIter              *child);
+static void               xfce_appfinder_category_category_free               (CategoryItem             *item);
 
 
 
@@ -73,12 +86,26 @@ struct _XfceAppfinderCategoryModelClass
 
 struct _XfceAppfinderCategoryModel
 {
-  GObject              __parent__;
-  gint                 stamp;
+  GObject                __parent__;
+  gint                   stamp;
 
-  GSList              *categories;
+  GSList                *categories;
 
-  GarconMenuDirectory *all_applications;
+  GarconMenuDirectory   *all_applications;
+
+  XfceAppfinderIconSize  icon_size;
+};
+
+struct _CategoryItem
+{
+  GarconMenuDirectory *directory;
+  GdkPixbuf           *pixbuf;
+};
+
+enum
+{
+  PROP_0,
+  PROP_ICON_SIZE
 };
 
 
@@ -94,7 +121,17 @@ xfce_appfinder_category_model_class_init (XfceAppfinderCategoryModelClass *klass
   GObjectClass *gobject_class;
 
   gobject_class = G_OBJECT_CLASS (klass);
+  gobject_class->get_property = xfce_appfinder_category_model_get_property;
+  gobject_class->set_property = xfce_appfinder_category_model_set_property;
   gobject_class->finalize = xfce_appfinder_category_model_finalize;
+
+  g_object_class_install_property (gobject_class,
+                                   PROP_ICON_SIZE,
+                                   g_param_spec_uint ("icon-size", NULL, NULL,
+                                                      XFCE_APPFINDER_ICON_SIZE_SMALLEST,
+                                                      XFCE_APPFINDER_ICON_SIZE_LARGEST,
+                                                      XFCE_APPFINDER_ICON_SIZE_DEFAULT_CATEGORY,
+                                                      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 }
 
 
@@ -104,9 +141,11 @@ xfce_appfinder_category_model_init (XfceAppfinderCategoryModel *model)
 {
   /* generate a unique stamp */
   model->stamp = g_random_int ();
+  model->icon_size = XFCE_APPFINDER_ICON_SIZE_DEFAULT_CATEGORY;
   model->all_applications = g_object_new (GARCON_TYPE_MENU_DIRECTORY,
                                           "name", _("All Applications"),
                                           "icon-name", "applications-other", NULL);
+  appfinder_refcount_debug_add (G_OBJECT (model->all_applications), "all-applications");
 }
 
 
@@ -130,16 +169,65 @@ xfce_appfinder_category_model_tree_model_init (GtkTreeModelIface *iface)
 
 
 
+static void
+xfce_appfinder_category_model_get_property (GObject      *object,
+                                            guint         prop_id,
+                                            GValue       *value,
+                                            GParamSpec   *pspec)
+{
+  XfceAppfinderCategoryModel *model = XFCE_APPFINDER_CATEGORY_MODEL (object);
+
+  switch (prop_id)
+    {
+    case PROP_ICON_SIZE:
+      g_value_set_uint (value, model->icon_size);
+      break;
+
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+    }
+}
+
+
+
+static void
+xfce_appfinder_category_model_set_property (GObject      *object,
+                                            guint         prop_id,
+                                            const GValue *value,
+                                            GParamSpec   *pspec)
+{
+  XfceAppfinderCategoryModel *model = XFCE_APPFINDER_CATEGORY_MODEL (object);
+  XfceAppfinderIconSize       icon_size;
+
+  switch (prop_id)
+    {
+    case PROP_ICON_SIZE:
+      icon_size = g_value_get_uint (value);
+      if (model->icon_size != icon_size)
+        {
+          model->icon_size = icon_size;
+
+          /* trigger a theme change to reload icons */
+          xfce_appfinder_category_model_icon_theme_changed (model);
+        }
+      break;
+
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+    }
+}
+
+
+
 static void
 xfce_appfinder_category_model_finalize (GObject *object)
 {
   XfceAppfinderCategoryModel *model = XFCE_APPFINDER_CATEGORY_MODEL (object);
-  GSList                     *li;
-  guint                       n;
 
-  /* clear the first three items */
-  for (li = model->categories, n = 0; li != NULL && n < 3; li = li->next, n++)
-    xfce_appfinder_model_category_free (li->data);
+  /* remove the categories */
+  g_slist_foreach (model->categories, (GFunc) xfce_appfinder_category_category_free, NULL);
   g_slist_free (model->categories);
 
   g_object_unref (G_OBJECT (model->all_applications));
@@ -257,7 +345,7 @@ xfce_appfinder_category_model_get_value (GtkTreeModel *tree_model,
           && item->directory != NULL)
         {
           icon_name = garcon_menu_directory_get_icon_name (item->directory);
-          item->pixbuf = xfce_appfinder_model_load_pixbuf (icon_name, 24);
+          item->pixbuf = xfce_appfinder_model_load_pixbuf (icon_name, model->icon_size);
         }
 
       g_value_init (value, GDK_TYPE_PIXBUF);
@@ -367,10 +455,26 @@ xfce_appfinder_category_model_iter_parent (GtkTreeModel *tree_model,
 
 
 
+
+static void
+xfce_appfinder_category_category_free (CategoryItem *item)
+{
+  if (item->directory != NULL)
+    g_object_unref (G_OBJECT (item->directory));
+  if (item->pixbuf != NULL)
+    g_object_unref (G_OBJECT (item->pixbuf));
+  g_slice_free (CategoryItem, item);
+}
+
+
+
+
 XfceAppfinderCategoryModel *
 xfce_appfinder_category_model_new (void)
 {
-  return g_object_new (XFCE_TYPE_APPFINDER_CATEGORY_MODEL, NULL);
+  gpointer model = g_object_new (XFCE_TYPE_APPFINDER_CATEGORY_MODEL, NULL);
+  appfinder_refcount_debug_add (G_OBJECT (model), "category-model");
+  return model;
 }
 
 
@@ -383,7 +487,7 @@ xfce_appfinder_category_model_set_categories (XfceAppfinderCategoryModel *model,
   GSList       *li, *lnext;
   GtkTreePath  *path;
   GtkTreeIter   iter;
-  guint         children, i;
+  guint         children;
 
   APPFINDER_DEBUG ("insert %d categories", g_slist_length (categories));
 
@@ -401,15 +505,21 @@ xfce_appfinder_category_model_set_categories (XfceAppfinderCategoryModel *model,
         }
      gtk_tree_path_free (path);
 
-     /* clean the first three categories and drop the list */
-     for (i = 0, li = model->categories; i < 3 && li != NULL; i++, li = li->next)
-       xfce_appfinder_model_category_free (li->data);
+     /* cleanup structures */
+     g_slist_foreach (model->categories, (GFunc)xfce_appfinder_category_category_free, NULL);
      g_slist_free (model->categories);
      model->categories = NULL;
     }
 
   appfinder_assert (model->categories == NULL);
 
+  for (li = categories; li != NULL; li = li->next)
+    {
+      item = g_slice_new0 (CategoryItem);
+      item->directory = g_object_ref (li->data);
+      model->categories = g_slist_prepend (model->categories, item);
+    }
+
   /* separator and the main categories */
   item = g_slice_new0 (CategoryItem);
   model->categories = g_slist_prepend (model->categories, item);
@@ -419,11 +529,12 @@ xfce_appfinder_category_model_set_categories (XfceAppfinderCategoryModel *model,
   model->categories = g_slist_prepend (model->categories, item);
 
   item = g_slice_new0 (CategoryItem);
-  item->directory = g_object_ref (G_OBJECT (model->all_applications));
+  item->directory = xfce_appfinder_model_get_bookmarks_category ();
   model->categories = g_slist_prepend (model->categories, item);
 
-  /* move the categories online */
-  model->categories = g_slist_concat (model->categories, g_slist_copy (categories));
+  item = g_slice_new0 (CategoryItem);
+  item->directory = g_object_ref (G_OBJECT (model->all_applications));
+  model->categories = g_slist_prepend (model->categories, item);
 
   path = gtk_tree_path_new_first ();
   for (li = model->categories; li != NULL; li = lnext)
@@ -452,11 +563,18 @@ xfce_appfinder_category_model_row_separator_func (GtkTreeModel *tree_model,
                                                   GtkTreeIter  *iter,
                                                   gpointer      user_data)
 {
-  CategoryItem *item = ITER_GET_DATA (iter);
+  CategoryItem *item;
 
   appfinder_return_val_if_fail (XFCE_IS_APPFINDER_CATEGORY_MODEL (tree_model), FALSE);
+  appfinder_return_val_if_fail (iter->stamp == XFCE_APPFINDER_CATEGORY_MODEL (tree_model)->stamp, FALSE);
 
-  return (item->directory == NULL);
+  if (iter->user_data != NULL)
+    {
+      item = ITER_GET_DATA (iter);
+      return (item->directory == NULL);
+    }
+
+  return FALSE;
 }
 
 
@@ -475,6 +593,7 @@ xfce_appfinder_category_model_icon_theme_changed (XfceAppfinderCategoryModel *mo
   for (li = model->categories, idx = 0; li != NULL; li = li->next, idx++)
     {
       item = li->data;
+      appfinder_assert (item != NULL);
 
       if (item->pixbuf != NULL)
         {
