@@ -42,7 +42,7 @@
 
 #ifdef GDK_WINDOWING_X11
 #include <gdk/gdkx.h>
-#define APPFINDER_WIDGET_XID(widget) ((guint) GDK_WINDOW_XID (GDK_WINDOW (GTK_WIDGET (widget)->window)))
+#define APPFINDER_WIDGET_XID(widget) ((guint) GDK_WINDOW_XID (gtk_widget_get_window (GTK_WIDGET (widget))))
 #else
 #define APPFINDER_WIDGET_XID(widget) (0)
 #endif
@@ -108,7 +108,9 @@ static gboolean   xfce_appfinder_window_item_visible                  (GtkTreeMo
 static void       xfce_appfinder_window_item_changed                  (XfceAppfinderWindow         *window);
 static void       xfce_appfinder_window_row_activated                 (XfceAppfinderWindow         *window);
 static void       xfce_appfinder_window_icon_theme_changed            (XfceAppfinderWindow         *window);
-static void       xfce_appfinder_window_execute                       (XfceAppfinderWindow         *window);
+static void       xfce_appfinder_window_launch_clicked                (XfceAppfinderWindow         *window);
+static void       xfce_appfinder_window_execute                       (XfceAppfinderWindow         *window,
+                                                                       gboolean                     close_on_succeed);
 
 
 
@@ -221,16 +223,27 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
   integer = xfconf_channel_get_int (window->channel, "/last/window-width", DEFAULT_WINDOW_WIDTH);
   gtk_window_set_default_size (GTK_WINDOW (window), integer, -1);
   gtk_window_set_icon_name (GTK_WINDOW (window), GTK_STOCK_EXECUTE);
+#if GTK_CHECK_VERSION (3, 0, 0)
+  gtk_window_set_has_resize_grip (GTK_WINDOW (window), FALSE);
+#endif
 
   if (xfconf_channel_get_bool (window->channel, "/always-center", FALSE))
     gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER);
 
+#if GTK_CHECK_VERSION (3, 0, 0)
+  vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 6);
+#else
   vbox = gtk_vbox_new (FALSE, 6);
+#endif
   gtk_container_add (GTK_CONTAINER (window), vbox);
   gtk_container_set_border_width (GTK_CONTAINER (vbox), 6);
   gtk_widget_show (vbox);
 
+#if GTK_CHECK_VERSION (3, 0, 0)
+  hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
+#else
   hbox = gtk_hbox_new (FALSE, 6);
+#endif
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, TRUE, 0);
   gtk_widget_show (hbox);
 
@@ -244,7 +257,11 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
   gtk_container_add (GTK_CONTAINER (align), image);
   gtk_widget_show (image);
 
+#if GTK_CHECK_VERSION (3, 0, 0)
+  vbox2 = gtk_box_new (GTK_ORIENTATION_VERTICAL, 6);
+#else
   vbox2 = gtk_vbox_new (FALSE, 6);
+#endif
   gtk_box_pack_start (GTK_BOX (hbox), vbox2, TRUE, TRUE, 0);
   gtk_widget_show (vbox2);
 
@@ -284,7 +301,11 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
   gtk_box_pack_start (GTK_BOX (vbox2), window->bin_collapsed, FALSE, TRUE, 0);
   gtk_widget_show (window->bin_collapsed);
 
+#if GTK_CHECK_VERSION (3, 0, 0)
+  window->paned = pane = gtk_paned_new (GTK_ORIENTATION_HORIZONTAL);
+#else
   window->paned = pane = gtk_hpaned_new ();
+#endif
   gtk_box_pack_start (GTK_BOX (vbox), pane, TRUE, TRUE, 0);
   integer = xfconf_channel_get_int (window->channel, "/last/pane-position", DEFAULT_PANED_POSITION);
   gtk_paned_set_position (GTK_PANED (pane), integer);
@@ -341,7 +362,11 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
   gtk_box_pack_start (GTK_BOX (vbox), window->bin_expanded, FALSE, TRUE, 0);
   gtk_widget_show (window->bin_expanded);
 
+#if GTK_CHECK_VERSION (3, 0, 0)
+  window->bbox = hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
+#else
   window->bbox = hbox = gtk_hbox_new (FALSE, 6);
+#endif
   gtk_widget_show (hbox);
 
   window->button_preferences = button = gtk_button_new_from_stock (GTK_STOCK_PREFERENCES);
@@ -349,9 +374,14 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
   g_signal_connect (G_OBJECT (button), "clicked",
       G_CALLBACK (xfce_appfinder_window_preferences), window);
 
+#if GTK_CHECK_VERSION (3, 0, 0)
+  bbox = gtk_button_box_new (GTK_ORIENTATION_HORIZONTAL);
+  gtk_box_set_spacing (GTK_BOX (bbox), 6);
+#else
   bbox = gtk_hbutton_box_new ();
-  gtk_button_box_set_layout (GTK_BUTTON_BOX (bbox), GTK_BUTTONBOX_END);
   gtk_button_box_set_spacing (GTK_BUTTON_BOX (bbox), 6);
+#endif
+  gtk_button_box_set_layout (GTK_BUTTON_BOX (bbox), GTK_BUTTONBOX_END);
   gtk_box_pack_start (GTK_BOX (hbox), bbox, TRUE, TRUE, 0);
   gtk_widget_show (bbox);
 
@@ -364,7 +394,7 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
   window->button_launch = button = gtk_button_new_with_mnemonic (_("La_unch"));
   gtk_container_add (GTK_CONTAINER (bbox), button);
   g_signal_connect_swapped (G_OBJECT (button), "clicked",
-      G_CALLBACK (xfce_appfinder_window_execute), window);
+      G_CALLBACK (xfce_appfinder_window_launch_clicked), window);
   gtk_widget_set_sensitive (button, FALSE);
   gtk_widget_show (button);
 
@@ -457,7 +487,7 @@ xfce_appfinder_window_key_press_event (GtkWidget   *widget,
   GtkWidget             *entry;
   XfceAppfinderIconSize  icon_size = XFCE_APPFINDER_ICON_SIZE_DEFAULT_ITEM;
 
-  if (event->keyval == GDK_Escape)
+  if (event->keyval == GDK_KEY_Escape)
     {
       gtk_widget_destroy (widget);
       return TRUE;
@@ -466,7 +496,7 @@ xfce_appfinder_window_key_press_event (GtkWidget   *widget,
     {
       switch (event->keyval)
         {
-        case GDK_l:
+        case GDK_KEY_l:
           entry = XFCE_APPFINDER_WINDOW (widget)->entry;
 
           gtk_widget_grab_focus (entry);
@@ -474,19 +504,19 @@ xfce_appfinder_window_key_press_event (GtkWidget   *widget,
 
           return TRUE;
 
-        case GDK_1:
-        case GDK_2:
+        case GDK_KEY_1:
+        case GDK_KEY_2:
           /* toggle between icon and tree view */
           xfconf_channel_set_bool (window->channel, "/icon-view",
-                                   event->keyval == GDK_1);
+                                   event->keyval == GDK_KEY_1);
           return TRUE;
 
-        case GDK_plus:
-        case GDK_minus:
-        case GDK_KP_Add:
-        case GDK_KP_Subtract:
+        case GDK_KEY_plus:
+        case GDK_KEY_minus:
+        case GDK_KEY_KP_Add:
+        case GDK_KEY_KP_Subtract:
           g_object_get (G_OBJECT (window->model), "icon-size", &icon_size, NULL);
-          if ((event->keyval == GDK_plus || event->keyval == GDK_KP_Add))
+          if ((event->keyval == GDK_KEY_plus || event->keyval == GDK_KEY_KP_Add))
             {
               if (icon_size < XFCE_APPFINDER_ICON_SIZE_LARGEST)
                 icon_size++;
@@ -496,8 +526,8 @@ xfce_appfinder_window_key_press_event (GtkWidget   *widget,
               icon_size--;
             }
 
-        case GDK_0:
-        case GDK_KP_0:
+        case GDK_KEY_0:
+        case GDK_KEY_KP_0:
           g_object_set (G_OBJECT (window->model), "icon-size", icon_size, NULL);
           return TRUE;
         }
@@ -782,6 +812,51 @@ xfce_appfinder_window_view_get_selected (XfceAppfinderWindow  *window,
 
 
 
+static void
+xfce_appfinder_window_popup_menu_toggle_bookmark (GtkWidget           *mi,
+                                                  XfceAppfinderWindow *window)
+{
+  const gchar  *uri;
+  GFile        *gfile;
+  gchar        *desktop_id;
+  GtkWidget    *menu = gtk_widget_get_parent (mi);
+  GtkTreeModel *filter;
+  GtkTreeModel *model;
+  GError       *error = NULL;
+
+  uri = g_object_get_data (G_OBJECT (menu), "uri");
+  if (uri != NULL)
+    {
+      gfile = g_file_new_for_uri (uri);
+      desktop_id = g_file_get_basename (gfile);
+      g_object_unref (G_OBJECT (gfile));
+
+      /* toggle bookmarks */
+      filter = g_object_get_data (G_OBJECT (menu), "model");
+      model = gtk_tree_model_filter_get_model (GTK_TREE_MODEL_FILTER (filter));
+      xfce_appfinder_model_bookmark_toggle (XFCE_APPFINDER_MODEL (model), desktop_id, &error);
+
+      g_free (desktop_id);
+
+      if (G_UNLIKELY (error != NULL))
+        {
+          g_printerr ("%s: failed to save bookmarks: %s\n", G_LOG_DOMAIN, error->message);
+          g_error_free (error);
+        }
+    }
+}
+
+
+
+static void
+xfce_appfinder_window_popup_menu_execute (GtkWidget           *mi,
+                                          XfceAppfinderWindow *window)
+{
+  xfce_appfinder_window_execute (window, FALSE);
+}
+
+
+
 static void
 xfce_appfinder_window_popup_menu_edit (GtkWidget           *mi,
                                        XfceAppfinderWindow *window)
@@ -921,14 +996,18 @@ xfce_appfinder_window_popup_menu (GtkWidget           *view,
   gchar        *title;
   gchar        *uri;
   GtkWidget    *mi;
+  GtkWidget    *image;
   gchar        *path;
   gboolean      uri_is_local;
+  gboolean      is_bookmark;
 
   if (xfce_appfinder_window_view_get_selected (window, &model, &iter))
     {
       gtk_tree_model_get (model, &iter,
                           XFCE_APPFINDER_MODEL_COLUMN_TITLE, &title,
-                          XFCE_APPFINDER_MODEL_COLUMN_URI, &uri, -1);
+                          XFCE_APPFINDER_MODEL_COLUMN_URI, &uri,
+                          XFCE_APPFINDER_MODEL_COLUMN_BOOKMARK, &is_bookmark,
+                          -1);
 
       /* custom command don't have an uri */
       if (uri == NULL)
@@ -942,6 +1021,7 @@ xfce_appfinder_window_popup_menu (GtkWidget           *view,
       menu = gtk_menu_new ();
       g_object_set_data_full (G_OBJECT (menu), "uri", uri, g_free);
       g_object_set_data_full (G_OBJECT (menu), "name", title, g_free);
+      g_object_set_data (G_OBJECT (menu), "model", model);
       g_signal_connect (G_OBJECT (menu), "selection-done",
           G_CALLBACK (gtk_widget_destroy), NULL);
 
@@ -954,6 +1034,28 @@ xfce_appfinder_window_popup_menu (GtkWidget           *view,
       gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
       gtk_widget_show (mi);
 
+      mi = gtk_image_menu_item_new_with_mnemonic (is_bookmark ? _("Remove From Bookmarks") : _("Add to Bookmarks"));
+      gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
+      g_signal_connect (G_OBJECT (mi), "activate",
+          G_CALLBACK (xfce_appfinder_window_popup_menu_toggle_bookmark), window);
+      gtk_widget_show (mi);
+
+      if (is_bookmark)
+        image = gtk_image_new_from_stock (GTK_STOCK_REMOVE, GTK_ICON_SIZE_MENU);
+      else
+        image = gtk_image_new_from_icon_name ("bookmark-new", GTK_ICON_SIZE_MENU);
+      gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mi), image);
+
+      mi = gtk_separator_menu_item_new ();
+      gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
+      gtk_widget_show (mi);
+
+      mi = gtk_image_menu_item_new_from_stock (GTK_STOCK_EXECUTE, NULL);
+      gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
+      g_signal_connect (G_OBJECT (mi), "activate",
+          G_CALLBACK (xfce_appfinder_window_popup_menu_execute), window);
+      gtk_widget_show (mi);
+
       mi = gtk_image_menu_item_new_from_stock (GTK_STOCK_EDIT, NULL);
       gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
       g_signal_connect (G_OBJECT (mi), "activate",
@@ -1006,10 +1108,12 @@ static void
 xfce_appfinder_window_set_padding (GtkWidget *entry,
                                    GtkWidget *align)
 {
-  gint padding;
+  gint          padding;
+  GtkAllocation alloc;
 
   /* 48 is the icon size of XFCE_APPFINDER_ICON_SIZE_48 */
-  padding = (48 - entry->allocation.height) / 2;
+  gtk_widget_get_allocation (entry, &alloc);
+  padding = (48 - alloc.height) / 2;
   gtk_alignment_set_padding (GTK_ALIGNMENT (align), MAX (0, padding), 0, 0, 0);
 }
 
@@ -1142,7 +1246,7 @@ xfce_appfinder_window_entry_activate (GtkEditable         *entry,
       if (cursor_set)
         gtk_widget_grab_focus (window->view);
       else
-        xfce_appfinder_window_execute (window);
+        xfce_appfinder_window_execute (window, TRUE);
     }
   else if (gtk_widget_get_sensitive (window->button_launch))
     {
@@ -1152,28 +1256,74 @@ xfce_appfinder_window_entry_activate (GtkEditable         *entry,
 
 
 
+static gboolean
+xfce_appfinder_window_pointer_is_grabbed (GtkWidget *widget)
+{
+#if GTK_CHECK_VERSION (3, 0, 0)
+  GdkDeviceManager *device_manager;
+  GList            *devices, *li;
+  GdkDisplay       *display;
+  gboolean          is_grabbed = FALSE;
+
+  display = gtk_widget_get_display (widget);
+  device_manager = gdk_display_get_device_manager (display);
+  devices = gdk_device_manager_list_devices (device_manager, GDK_DEVICE_TYPE_MASTER);
+
+  for (li = devices; li != NULL; li = li->next)
+    {
+      if (gdk_device_get_source (li->data) == GDK_SOURCE_MOUSE
+          && gdk_display_device_is_grabbed (display, li->data))
+        {
+          is_grabbed = TRUE;
+          break;
+        }
+    }
+
+  g_list_free (devices);
+
+  return is_grabbed;
+#else
+  return gdk_pointer_is_grabbed ();
+#endif
+}
+
+
+
 static gboolean
 xfce_appfinder_window_entry_key_press_event (GtkWidget           *entry,
                                              GdkEventKey         *event,
                                              XfceAppfinderWindow *window)
 {
-  gboolean expand, is_expanded;
+  gboolean          expand;
+  gboolean          is_expanded;
 
-  if (event->keyval == GDK_Up || event->keyval == GDK_Down)
+  if (event->keyval == GDK_KEY_Up
+      || event->keyval == GDK_KEY_Down)
     {
-      expand = (event->keyval == GDK_Down);
+      expand = (event->keyval == GDK_KEY_Down);
       is_expanded = gtk_widget_get_visible (window->paned);
       if (is_expanded != expand)
         {
           /* don't break entry completion navigation in collapsed mode */
           if (!is_expanded
-              && gdk_pointer_is_grabbed ())
-            return FALSE;
+              && xfce_appfinder_window_pointer_is_grabbed (entry))
+            {
+              /* window is still collapsed and the pointer is grabbed
+               * by the popup menu, do nothing with the event */
+              return FALSE;
+            }
 
           xfce_appfinder_window_set_expanded (window, expand);
           return TRUE;
         }
     }
+  else if (event->keyval == GDK_KEY_Tab
+           && !gtk_widget_get_visible (window->paned)
+           && xfce_appfinder_window_pointer_is_grabbed (entry))
+    {
+      /* don't tab to the close button */
+      return TRUE;
+    }
 
   return FALSE;
 }
@@ -1236,7 +1386,7 @@ xfce_appfinder_window_treeview_key_press_event (GtkWidget           *widget,
 {
   if (widget == window->view)
     {
-      if (event->keyval == GDK_Left)
+      if (event->keyval == GDK_KEY_Left)
         {
           gtk_widget_grab_focus (window->sidepane);
           return TRUE;
@@ -1244,7 +1394,7 @@ xfce_appfinder_window_treeview_key_press_event (GtkWidget           *widget,
     }
   else if (widget == window->sidepane)
     {
-      if (event->keyval == GDK_Right)
+      if (event->keyval == GDK_KEY_Right)
         {
           gtk_widget_grab_focus (window->view);
           return TRUE;
@@ -1407,7 +1557,8 @@ xfce_appfinder_window_item_visible (GtkTreeModel *model,
   XfceAppfinderWindow *window = XFCE_APPFINDER_WINDOW (data);
 
   return xfce_appfinder_model_get_visible (XFCE_APPFINDER_MODEL (model), iter,
-                                           window->filter_category, window->filter_text);
+                                           window->filter_category,
+                                           window->filter_text);
 }
 
 
@@ -1524,7 +1675,16 @@ xfce_appfinder_window_execute_command (const gchar          *text,
 
 
 static void
-xfce_appfinder_window_execute (XfceAppfinderWindow *window)
+xfce_appfinder_window_launch_clicked (XfceAppfinderWindow *window)
+{
+  xfce_appfinder_window_execute (window, TRUE);
+}
+
+
+
+static void
+xfce_appfinder_window_execute (XfceAppfinderWindow *window,
+                               gboolean             close_on_succeed)
 {
   GtkTreeModel *model;
   GtkTreeIter   iter, orig;
@@ -1592,7 +1752,7 @@ xfce_appfinder_window_execute (XfceAppfinderWindow *window)
       g_error_free (error);
     }
 
-  if (result)
+  if (result && close_on_succeed)
     gtk_widget_destroy (GTK_WIDGET (window));
 }
 
